#!/usr/bin/perl -w

# =============================================================================
# FILE:          perlAPRS
#
# VERSION:       1.0.3
#
# AUTHOR:        Richard Parry, W9IF
#
# DESCRIPTION:   This perl program will read a command file that
# provides a callsign, command, grid sqaure, and command count.
# The program will then listen to an APRS port and execute the command
# script if that station is heard in the grid square specified.
#
# SOURCE:        See perlAPRS web page for latest version and further
# details of how to use perlAPRS
#
#                http://people.qualcomm.com/rparry
#
# Type the following command to get short help menu:
#
#                perlAPRS -h
#
# =============================================================================


require 5.002;


@APRSsigns = ("AIR", "ALL", "AP", "ID", "CQ", "QST", "BEACON", "MAIL", "SKY",
	      "GPS", "DFNET", "SPCL", "TEST", "DRILL", "AIR", "DF", "DGPS", 
		   "DRILL", "DX", "JAVA", "MICE", "QTH", "RTCM", "SPC", "SYM", "TEL",
		   "TEST", "TLM", "WX", "ZIP"  );

# These arrays of hashes (associative array) are used to ease the computation
# of the grid squares.  Since lat and lon are in mixed format containing
# Degrees, Minutes, and Decimal Minutes, the computation is not
# straight forward.

%gridchar4 = (
		 A => {deg => '0', min => '00',},
		 B => {deg => '1', min => '55',},
		 C => {deg => '1', min => '50',},
		 D => {deg => '1', min => '45',},
		 E => {deg => '1', min => '40',},
		 F => {deg => '1', min => '35',},
		 G => {deg => '1', min => '30',},
		 H => {deg => '1', min => '25',},
		 I => {deg => '1', min => '20',},
		 J => {deg => '1', min => '15',},
		 K => {deg => '1', min => '10',},
		 L => {deg => '1', min => '05',},
		 M => {deg => '1', min => '00',},
		 N => {deg => '2', min => '55',},
		 O => {deg => '2', min => '50',},
		 P => {deg => '2', min => '45',},
		 Q => {deg => '2', min => '40',},
		 R => {deg => '2', min => '35',},
		 S => {deg => '2', min => '30',},
		 T => {deg => '2', min => '25',},
		 U => {deg => '2', min => '20',},
		 V => {deg => '2', min => '15',},
		 W => {deg => '2', min => '10',},
		 X => {deg => '2', min => '05',},
		 Y => {deg => '2', min => '00',},
		 );

%gridchar5 = (
		 A => {deg => '0', min => '00', decmin => '.0'},
		 B => {deg => '0', min => '02', decmin => '.5'},
		 C => {deg => '0', min => '05', decmin => '.0'},
		 D => {deg => '0', min => '07', decmin => '.5'},
		 E => {deg => '0', min => '10', decmin => '.0'},
		 F => {deg => '0', min => '12', decmin => '.5'},
		 G => {deg => '0', min => '15', decmin => '.0'},
		 H => {deg => '0', min => '17', decmin => '.5'},
		 I => {deg => '0', min => '20', decmin => '.0'},
		 J => {deg => '0', min => '22', decmin => '.5'},
		 K => {deg => '0', min => '25', decmin => '.0'},
		 L => {deg => '0', min => '27', decmin => '.5'},
		 M => {deg => '0', min => '30', decmin => '.0'},
		 N => {deg => '0', min => '32', decmin => '.5'},
		 O => {deg => '0', min => '35', decmin => '.0'},
		 P => {deg => '0', min => '37', decmin => '.5'},
		 Q => {deg => '0', min => '40', decmin => '.0'},
		 R => {deg => '0', min => '42', decmin => '.5'},
		 S => {deg => '0', min => '45', decmin => '.0'},
		 T => {deg => '0', min => '47', decmin => '.5'},
		 U => {deg => '0', min => '50', decmin => '.0'},
		 V => {deg => '0', min => '52', decmin => '.5'},
		 W => {deg => '0', min => '55', decmin => '.0'},
		 X => {deg => '0', min => '57', decmin => '.5'},
		 Y => {deg => '0', min => '60', decmin => '.0'},
		 );

# Fixed subscripts into the database array "cinput"
# The first 4 values are read in from the user's command file.
$ptrgridsquare   = 1;
$ptrlatlow       = 2;
$ptrlonlow       = 3;
$ptrlatupr       = 4;
$ptrlonupr       = 5;


# -----------------------------------------------------------------------------
# Global Variable Declarations and Initialization
# -----------------------------------------------------------------------------





# =============================================================================
# MAIN PROGRAM
# =============================================================================



sub initialise {
	@command = @_;
	$cmdcount = 0;
	foreach $com (@command){
    	$com =~ s/\W//;
    	$cinput[$cmdcount][$ptrgridsquare]= $com;
    	$cmdcount++;
	}

	
	if ($show) {
		print "\n\t\t\t*** ACCEPT ***\n";
		print "     Grid  LwrLat  LwrLon    UprLat  UprLon\n";
	}

	for ($i = 0; $i < $cmdcount; $i++) {
   	@grid = unpack "cccccc", $cinput[$i][$ptrgridsquare];

		# For 2 Letter Grid Square
   	if (scalar @grid == 2) {
			$latlow = (($grid[1] - 65) * 10 - 90)  * 100;
			$lonlow = (($grid[0] - 65) * 20 - 180) * 100;
			$latupr = $latlow + 1000;
			$lonupr = $lonlow + 2000;
    	}
    
		# For 4 Letter Grid Square
		if (scalar @grid == 4) {
			$latlow = (($grid[1] - 65) * 10 + ($grid[3] - 48) * 1 - 90)  * 100;
			$lonlow = (($grid[0] - 65) * 20 + ($grid[2] - 48) * 2 - 180) * 100;
			$latupr = $latlow + 100;
			$lonupr = $lonlow + 200;
    	}
    
		# For 6 Letter Grid Square format "DegreesMinutes.DecimalMinutes"
		if (scalar @grid == 6) {
			# compute latitude of lower corner
			$latdegrees = ($grid[1] - 65) * 10 + ($grid[3] - 48) - 90 +
	    		$gridchar5{chr($grid[5])}{deg};
			$latlow = pack "a*a*a*", $latdegrees, $gridchar5{chr($grid[5])}{min},
	    		$gridchar5{chr($grid[5])}{decmin};
	
			# compute longitude of lower corner
			$londegrees = ($grid[0] - 65) * 20 + ($grid[2] - 48) * 2 - 180 +
	    		$gridchar4{chr($grid[4])}{deg};
			$lonlow = pack "a*a*", $londegrees, $gridchar4{chr($grid[4])}{min};
	
			# compute latitude of upper corner
			$char5 = chr($grid[5] + 1);       # increment character
			$latdegrees = ($grid[1] - 65) * 10 + ($grid[3] - 48) - 90 + 
	    		$gridchar5{$char5}{deg};
			$latupr = pack "a*a*a*", $latdegrees, $gridchar5{$char5}{min},
	    		$gridchar5{$char5}{decmin};
	
			# compute longitude of upper corner 
			$char4 = chr($grid[4] + 1);       # increment character
			$londegrees = ($grid[0] - 65) * 20 + ($grid[2] - 48) * 2 - 180 +
	    		$gridchar4{$char4}{deg};
			$lonupr = pack "a*a*", $londegrees, $gridchar4{$char4}{min};
		}	
    
		$cinput[$i][$ptrlatlow] = $latlow;
   	$cinput[$i][$ptrlonlow] = $lonlow;
   	$cinput[$i][$ptrlatupr] = $latupr;
   	$cinput[$i][$ptrlonupr] = $lonupr;    
    
		# Print comparison array which contains all data for each callsign
    	if ($show) {
			printf "%02d %6s %8.1f %8.1f %8.1f %8.1f\n",
			$i + 1,
			$cinput[$i][$ptrgridsquare],
			$cinput[$i][$ptrlatlow],
			$cinput[$i][$ptrlonlow],
			$cinput[$i][$ptrlatupr],
			$cinput[$i][$ptrlonupr];
    	}	
	}
}




sub validate_line { 
	local ($_) = shift;
     chomp;                                    # remove newline char
     my $packet = $_;                             # save the packet
     if (length $packet eq 0) { return (""); };   # skip all if a null packet
     if ($show) {
			 print "Packet= $packet\n";            # print input line
     }

#     
# Split up the packet and store elements into array
#
     ($fromcallsign) = split(/>/, $packet, 0); # split packet by ">" delimiter
     $fromcallsign =~ s/\*//;                  # remove any trailing * char
     $fromcallsign =~ s/^[^A-Z]//;             # remove leading non-alpha chars




#
# In this block, search for a valid "TO" address.  If none found set flag
# and exit block.  If found, set flag for valid APRS packet.  However,
# additional testing will be done later to confirm packet contains valid
# latitude and longitude values.
# If the portion of the packet that we now have begins with a legitimate
# "TO ADDRESS", then accept this packet and proceed, but remove the 
# TO ADDRESS, it is no longer needed.
#
   APRSBLOCK: {
       $validaprs = 1;                        # assume legitimate APRS packet
       foreach $testcall (@APRSsigns) {
	   if ($packet =~ m/\>$testcall/) {   # search for leading ">" + sign
	       last APRSBLOCK;
	   }
       }
       $validaprs = 0;                        # set flag, not APRS packet
   } # END APRSBLOCK



#
# Extract Lat and Long if in form 99.99?N/99999.99?W
#                              or 99.99?N\99999.99?W
#                              or 99.99?N[A-Z]99999.99?W
#
   TNCBLOCK: {
       if($validaprs) {
	   $validtnc = 1;        # assume valid latitude or longitude
#	   if ($packet =~ m/([0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?[N|n][A-Z|\/][0-9][0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?[W|w])/)
	   if ($packet =~ m/([0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?[S|s|N|n].[0-9][0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?[W|w|E|e])/)
	   {
	       $latlong = $1;    # save latitude and longitude
	       $latlong =~ m/([0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?)/;
	       $lat = $1;
	       if ($latlong =~ m/([0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?[S|s])/){
		       $lat = -$lat;
			 }

	
	       $latlong =~ m/([0-9][0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?)/;
	       $long = $1;
	       if ($latlong =~ m/([0-9][0-9][0-9][0-9][0-9]\.[0-9][0-9][0-9]?[W|w])/){
	       		$long = -$long;
			 }
	       last TNCBLOCK;
	   }
	   $validtnc = 0;        # not valid latitude or longitude
       }
   } # END TNCBLOCK



#
# Now find out if it is a NMEA standard string
# Strings supported are:
#       $GPGGA Global Positioning System Fix Data
#       $GPGLL Geographic position, latitude and longitude
#       $GPRMC Recommended minimum specific GPS/Transit data
#
   NMEASWITCH: {
       if($validaprs && !$validtnc) {
	   $validnmea = 1;                    # assume valid NMEA packet
	   if ($packet  =~ m/\$GPGGA/)  {
	       # separate header and payload, we need only the payload)
	       ($header, $payload) = split(/\$GPGGA/, $packet);
#	       $payload =~ s/[A-Z]//g;        # remove chars in bad NMEAs
	       @posit = split(/,/, $payload); # disect $GP sentence into fields
			 if (index ($posit[2], "N") >= 0){			
	       	$lat   = $posit[2];            # extract latitude
	       } else {
				$lat   = -$posit[2];            # extract latitude
			 }
			 if (index ($posit[4], "E") >= 0){
		       $long  = $posit[4];            # extract longitude
			 } else {
		       $long  = -$posit[4];            # extract longitude
			 }
			 $lat =~  s/[A-Z]//g;
			 $long =~  s/[A-Z]//g;
	       if ($lat eq "" or $long eq "") { $validnmea = 0} # invalid NMEA
	       last NMEASWITCH;
	   }
	   
	   if ($packet  =~ m/\$GPGLL/)  {
	       # separate header and payload, we need only the payload)
	       ($header, $payload) = split(/\$GPGLL/, $packet);
#	       $payload =~ s/[A-Z]//g;        # remove chars in bad NMEAs
	       @posit = split(/,/, $payload); # disect $GP sentence into fields
			 if (index ($posit[1], "N") >= 0){			
	       	$lat   = $posit[1];            # extract latitude
	       } else {
				$lat   = -$posit[1];            # extract latitude
			 }
			 if (index ($posit[3], "E") >= 0){
		       $long  = $posit[3];            # extract longitude
			 } else {
		       $long  = -$posit[3];            # extract longitude
			 }
			 $lat =~  s/[A-Z]//g;
			 $long =~  s/[A-Z]//g;

	       if ($lat eq "" or $long eq "") { $validnmea = 0} # invalid NMEA
	       last NMEASWITCH;
	   }
	   
	   if ($packet  =~ m/\$GPRMC/)  {
	       # separate header and payload, we need only the payload)
	       ($header, $payload) = split(/\$GPRMC/, $packet);
#	       $payload =~ s/[A-Z]//g;        # remove chars in bad NMEAs
	       @posit = split(/,/, $payload); # disect $GP sentence into fields
			 if (index ($posit[3], "N") >= 0){			
	       	$lat   = $posit[3];            # extract latitude
	       } else {
				$lat   = -$posit[4];            # extract latitude
			 }
			 if (index ($posit[5], "E") >= 0){
		       $long  = $posit[5];            # extract longitude
			 } else {
		       $long  = -$posit[5];            # extract longitude
			 }
			 $lat =~  s/[A-Z]//g;
			 $long =~  s/[A-Z]//g;
	       if ($lat eq "" or $long eq "") { $validnmea = 0} # invalid NMEA
	       last NMEASWITCH;
	   }
	   $validnmea = 0;                    # Invalid NMEA packet
       }
   } # END NMEA BLOCK
     


     
#    
# We have valid APRS packet, check for match in command database, if so execute
#
     if ($validaprs && ($validtnc or $validnmea)) {
	 if ($show) {
	     printf "\t%11s       %4.3f    %5.3f\n", $fromcallsign, $lat, $long;
	 }

	 # Current version assumes station is in North American,
	 # N,S,E,W parameters ignored, therefore force to negative longitude
#	 $long = -$long;

	 # Is packet from within the gridsquare of any callsign in the database
	 for ($i = 0; $i < $cmdcount; $i++) {
	     if(($lat > $cinput[$i][$ptrlatlow]) &&
				($lat < $cinput[$i][$ptrlatupr]) &&
				($long > $cinput[$i][$ptrlonlow]) &&
				($long < $cinput[$i][$ptrlonupr])) {

				 # get current time to time stamp this packet
		 

				 if ($show) {
		    		 $cinput[$i][$ptrgridsquare];
		 		 }


				return ($cinput[$i][$ptrgridsquare]);
	   	  }


	 }


     }
     else
     {
	 if ($debug) {
	     print "*No Posit= $packet\n";
	 }
#	 next;
     }

#sleep 1;     # For testing only
     
 }

1;